/*
    This file is part of Frozen Bubble for S60.

    Frozen Bubble for S60 is free software; you can redistribute
    it and/or modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2 of
    the License, or (at your option) any later version.

    Frozen Bubble for S60 is distributed in the hope that it
    will be useful, but WITHOUT ANY WARRANTY; without even the implied
    warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
    See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Frozen Bubble for S60; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef CFROZENBUBBLEMAIN_H
#define CFROZENBUBBLEMAIN_H

#include <e32base.h>
#include <e32std.h>
#include <eikenv.h>
#include <e32keys.h>
#include "TSample.h"

// Forward declarations
class CExtendedBitmap;
class CPrintableExtendedBitmap;
class TShootData;
class CExtendedFont;
class TSample;
class CSndMixer;
class CWavLoader;

// blit types
const TInt BLTTYPE_NORMAL = 0;
const TInt BLTTYPE_MASKED = 1;

// External events
const TInt EVT_ESCAPE   = 1;
const TInt EVT_GOBACKGR = 2;
const TInt EVT_GOFOREGR = 3;
const TInt EVT_SELECT   = 4;

// Frozen Bubble interface for passing commands to FrozenBubbleAppUi
class MFrozenBubbleAction{
  public:
    // Graphic functions
    virtual void FlipBackBuffer(void)=0;
    virtual void BitBlt(TInt iBltType, CExtendedBitmap &iBitmap, const TPoint &aPosition)=0;
    virtual void BitBlt(TInt iBltType, CExtendedBitmap &iBitmap, const TPoint &aPosition, const TRect &aRect)=0;
    virtual void BitBltBack(CPrintableExtendedBitmap &iBitmapTarget)=0;
    virtual void DoExit(void)=0;
    virtual TBool ReadIniFile(TUid iInfo, void *ptr, TUint &size)=0;
    virtual TBool CheckIniFile(TUid iInfo)=0;
    virtual void WriteIniFile(TUid iInfo, void *ptr, TUint size)=0;
    virtual void RemoveIniData(TUid iInfo)=0;
};

const TInt CANON_ROTATIONS_NB = 40; // number of rotations of images for canon (should be consistent with gfx/shoot/Makefile)

// Highscores
const TInt HIGHSCORELIMIT = 10;
typedef char HighScoreName[HIGHSCORELIMIT];
class THighScoreRecord{
  public:
    TInt level;
    TInt seconds;
    HighScoreName name;
};

typedef RArray<THighScoreRecord> RHighScores;
typedef TLinearOrder<THighScoreRecord> THighScoreRecordLinearOrder;

// Basic Bubble information class
class TBubbleBase{
  public:
    TInt8 iColorIndex;
    TInt8 cx, cy;
    TBubbleBase(void);
//    TBubbleBase(const TBubbleBase& bubble);
//    ~TBubbleBase();
};

// Bubble information class
class TBubble:public TBubbleBase{
  public:
    TInt16 chaindestx, chaindesty;
    TInt16 chaindestcx, chaindestcy;
    TReal x, y;
    TReal direction;
    TReal x_old, y_old;
    TReal speedx, speedy;
    TReal speed;
    TBubbleBase neighbours[6];
    TInt8 stick_y;
    TInt8 mark;
    TBool chaingoingup;
    TInt wait_fall;
    TInt addNeighbour(const TBubbleBase &bubble);
    TBubble(void);
    TInt neighboursCount(void);
    void removeNeighbour(TInt idx);
};

typedef RArray<TBubble> RBubbleArray;
typedef TLinearOrder<TBubble> TBubbleLinearOrder;

// Frozen Bubble engine
class CFrozenBubbleMain:public CBase{
  private:
    TPoint ptZero;
    TSize iGameRes, iScreenRes;
    TTime total_time;
    TInt64 iRandomSeed;
    CPeriodic *prdGameLoop;
    TBool bFirstLoopIteration;
    TInt global_state;
    // Game state variables
    TStdScanCode KEYS_left, KEYS_right, KEYS_fire, KEYS_center;
    TStdScanCode SET_left, SET_right, SET_fire, SET_center;
    TBool bSoundEnabled, bLoadedGame;
    TInt pdataScore, pdataNewroot, pdataNewroot_prelight, pdataOddswap, pdataHurry, pdataNewrootlevel,
         pdataNewroot_prelight_step;
    TInt pdataPing_rightImg, pdataPing_rightState, pdataPing_rightMovelatency, pdataState, pdataHurry_oddness, 
         pdataSticking_step, pdataSleeping, pdataPing_rightActionlatency, pdataSticking_step_slowdown;
    TInt pinguinWin_roll_back_index, pinguinLose_roll_back_index;
    TBool actionsLeft, actionsRight, actionsFire, actionsCenter, actionsHadfire;
    TReal angle;
    RBubbleArray sticked_bubbles, falling_bubble, exploding_bubble,
                 chains_falling_chained, chains_chained_bubbles, tmp_bubbles;
    RBubbleArray will_destroy;
    TBubble launched_bubble, sticking_bubble;
    TBubble next_bubble, tobe_launched;
    // high score vars
    RHighScores highscores;
    THighScoreRecord candidateHighScore;
    TInt ticks_last_key;
    TStdScanCode sccPreviousKeyEvent;
    TInt hs_ypos;
    TBool editEnded;
    TUint time_1pgame, time_1pgameextra;
    // Pause vars
    TInt8 pauseIndex, pauseSide;
    // Game constants
    TInt POSleft_limit, POSright_limit, POSscoresx, POSinit_top_limit, POSinitial_bubble_y,
         POSmalus_y, POSscoresy, POScompressor_xpos, POStop_limit, BUBBLE_SIZE, ROW_SIZE;
    TPoint POSnext_bubble, POSpinguin, POShurry, POScenterpanel, POSpause_clip;
    TReal PI;
    // Level data
    TInt LevelsCurrent;
    TInt8 levels[100][10][8];
    // key variables
    TStdScanCode sccPressedButton;
    TStdScanCode sccLastKeyEvent;
    TInt iMenuCurrent_pos;
    // Menu variables
    TInt iBannerPos;
    TInt BANNERS[4];
    TInt BANNER_START, BANNER_SPACING, BANNERS_MAX;
    TInt BANNER_MINX, BANNER_MAXX, BANNER_Y;
    TRect BANNER_RECT;
    // Sounds
    TSample smpStick, smpLaunch, smpDestroy_group, smpApplause, smpCancel, smpHurry,
            smpLose, smpMenu_change, smpMenu_selected, smpNewroot_solo, smpNoh, 
            smpRebound, smpSnore, smpTypewriter;
    CSndMixer *smxMain;
    TInt8 iActiveChannel;
    // Fonts
    CExtendedFont *fntNormal, *fntDark;
    // --- Bitmap section ---
    CExtendedBitmap *imgBack_1p, *imgBackstartfull, *imgBack_paused, 
                    *imgLose, *imgWin, *imgVoidPanel, *imgBack_hiscores, 
                    *imgHurry/*, *imgQueryYesNo*/;
    CExtendedBitmap *imgPause[5];
    CPrintableExtendedBitmap *imgOrigBackground;
    // Menu bitmaps
    CExtendedBitmap *imgTxt_1pgame_off, *imgTxt_1pgame_over,
                    *imgTxt_keys_off, *imgTxt_keys_over,
                    *imgTxt_sound_off, *imgTxt_sound_over,
                    *imgTxt_sound_act_off, *imgTxt_sound_act_over,
                    *imgTxt_highscores_off, *imgTxt_highscores_over;
    // Pinguin bitmaps
    CExtendedBitmap *imgPinguinNormal[4];
    CExtendedBitmap *imgPinguinAction, *imgPinguinMoveLeft, *imgPinguinMoveRight, *imgPinguinSleep;
    CExtendedBitmap *imgPinguinLose[9], *imgPinguinWin[9];
    // Menu banners
    CExtendedBitmap *imgBannerArtwork, *imgBannerCpucontrol, *imgBannerPorting, *imgBannerSoundtrack;
    // Compressor bitmaps
    CExtendedBitmap *imgCompressorExt, *imgCompressorMain;                    
    // ball bitmaps
    CExtendedBitmap *imgBall[8];
    CExtendedBitmap /**imgBubblesAnimWhite, */*imgBubblesAnimLose;
    CExtendedBitmap *imgBubblesAnim[8];
    // canon
    CExtendedBitmap *imgCanon[2*CANON_ROTATIONS_NB+1];
    TShootData *canondata;
    // ----------------------
    MFrozenBubbleAction *Actions;
    CEikonEnv* EikonEnv;
    TBuf<80> iPrivatePath;
  public:
    CFrozenBubbleMain(MFrozenBubbleAction *ActionListener):iRandomSeed(getRandomSeed()){
      setConstants();
      Actions = ActionListener;
      EikonEnv = CEikonEnv::Static();
      iGameRes.SetSize(240, 320);
    }
    ~CFrozenBubbleMain();
    TKeyResponse DoKeyEvent(const TKeyEvent& aKeyEvent, TEventCode aType);
    void ExternalEvent(TInt event);
    void ForceQuit(void);
    void main(void);
    TSize GetGameScreenSize(void){
      return iGameRes;
    }
  private:
    // ported functions
    void init_game(void);
    void init_game_graphics(void);
    void init_game_sound(void);
    void load_sound_sample(CWavLoader& loader, TSample& sample, const TFileName& filename);
    void init_game_rest(void);
	void play_music(const TDesC8& aMusicName);

    void menu();
    TBool handle_new_hiscores();
    void ask_name(void);
    static TInt ask_nameloop_(TAny* aPtr);
    void ask_nameloop(void);
    void ask_keys(void);
    static TInt ask_keysloop_(TAny* aPtr);
    void ask_keysloop(void);
    void display_highscores(void);
    static TInt highscoreloop_(TAny* aPtr);
    void highscoreloop(void);
    TBool checksaved(void);
    void querysavegame(void);
    void dopause(void);
    static TInt dopause_loop_(TAny* aPtr);
    void dopause_loop(void);
    void new_game(TBool loadgame);
    static TInt maingame_(TAny* aPtr);
    void maingame(void);
    void update_game(void);
    void find_bubble_group(TBubble &b, RBubbleArray &group);
    void handle_game_events(void);
    void stick_bubble(TBubble &bubble, TInt xpos, TInt ypos, TBool count_for_root);
    void verify_if_end();
    void handle_graphics(TBool iPutImage);
    void print_compressor();
    void load_levelset();
    void print_scores();
    void real_stick_bubble(TBubble &bubble, TInt xpos, TInt ypos, TBool neighbours_ok);
    void open_level(TInt iLevel);
    void RebuildOriginalBackground(void);
    void remove_images_from_background(RBubbleArray &images);
    void erase_image(CExtendedBitmap &image, const TPoint &ptTarget);
    void put_allimages_to_background(RBubbleArray &ba);
    void difference2(RBubbleArray &barr, RBubbleArray &bdiff, RBubbleArray &bres);
    void destroy_bubbles(RBubbleArray &bubz);
    void destroy_bubbles(TBubble &bub);
    TBool is_collision(TBubble &bub, TReal x, TReal y);
    TPoint get_array_closest_pos(TInt x, TInt y);
    TInt get_array_yclosest(TReal y);
    TBool bubble_next_to(TInt x1, TInt y1, TInt x2, TInt y2);
    TBubble create_bubble_given_img(TInt8 iColor);
    TBubble create_bubble(TBool create_same_color);
    void print_hurry();
    void remove_hurry();
    void print_next_bubble(TInt8 iColor);
    void print_bubble(TBubble &bubble);
    void generate_new_bubble(TInt8 iColor);
    void calc_real_pos(TBubble &b);
    void calc_real_pos_given_arraypos(TInt8 cx, TInt8 cy, TReal &newx, TReal &newy);
    static TInt menu_loop_(TAny* aPtr);
    void menu_loop(void);
    void AppendNeighboursOf(RBubbleArray &bubbles, RBubbleArray &dest);
    TBool IsColorInBubbleArray(TInt8 iColor, RBubbleArray &rarray);
    TBool AnyExistsChaindestx(RBubbleArray &barray);
    TBool AnyChaindestx(RBubbleArray &barray);
    TInt FindBubbleInArray(TBubbleBase &b, RBubbleArray &br);
    CExtendedBitmap* StateToPinguinImage(TInt state, TInt index);
    TInt CompareBubbles(TBubbleBase &b1, TBubbleBase &b2);
    void play_sound(TSample& sample);
    // primitive functions
    void load_settings(void);
    void save_settings(void);
    void load_game(void);
    void save_game(void);
    void load_hiscores(void);
    void save_hiscores(void);
    static TInt BubblePositionOrder(const TBubble &a, const TBubble &b);
    static TInt BubbleCXOrder(const TBubble &a, const TBubble &b);
    static TInt HighScoreSort(const THighScoreRecord &a, const THighScoreRecord &b);
    CFbsBitmap* ConvertBitmapL(CFbsBitmap *original, TDisplayMode mode);
    void InitializeBitmap(CExtendedBitmap** iBmp, CFbsBitmap* iBitmap, TBool iDisposeBitmap, CFbsBitmap* iAlpha, TBool iDisposeAlpha);
    void AddBrightness(CPrintableExtendedBitmap &imgToBrighten);
    TUint getAppTicks(void);
    void InitTimer(void);
    void ResetTimeStamp(void);
    void CalcTimer(void);
    void setConstants(void);
    TInt64 getRandomSeed(void);
    void setGameLoop(TInt (*iFunction)(TAny*));
    TReal frand(TReal max=1.0);
    TInt rand(TInt dividor);
};

#endif
