/*
    This file is part of Frozen Bubble for S60.

    Frozen Bubble for S60 is free software; you can redistribute
    it and/or modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2 of
    the License, or (at your option) any later version.

    Frozen Bubble for S60 is distributed in the hope that it
    will be useful, but WITHOUT ANY WARRANTY; without even the implied
    warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
    See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Frozen Bubble for S60; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "CSndMixer.h"
#include "CMixerThread.h"
#include <e32svr.h>

_LIT(KMixer, "Mixer");


CSndMixer* CSndMixer::NewL() {
  CSndMixer* self = new( ELeave )CSndMixer();
  CleanupStack::PushL( self );
  self->ConstructL();
  CleanupStack::Pop( self );
  return self;
}


void CSndMixer::ConstructL() {
  iShared.iMainVolume = 160;//256;  // default main volume

  /* Set the MikMod music format */
  
  md_mode = DMODE_16BITS;
  md_mixfreq	 = 16000;
  md_device	  = 0;
  md_volume	  = 64;
  md_musicvolume = 64;
  md_sndfxvolume = 64;
  md_pansep	  = 128;
  md_reverb	  = 0;
  
  MikMod_RegisterAllDrivers();	
  MikMod_RegisterAllLoaders();
  if ( MikMod_Init(NULL) ) {
	  User::Leave(KErrNotSupported);
  } 

  User::LeaveIfError( iShared.iAliveMutex.CreateLocal( 1 ) );
  User::LeaveIfError( iShared.iMutex.CreateLocal() );
  User::LeaveIfError( iMixerThread.Create( KMixer,
        CMixerThread::ThreadFunction,
        KDefaultStackSize,
        NULL, 
        &iShared));
  
  // Give all possible priority to audio
  iMixerThread.SetPriority( EPriorityRealTime );
  iMixerThread.Resume();
  iPaused = ETrue;
}


CSndMixer::CSndMixer() {
}


CSndMixer::~CSndMixer() {
  SendCmd( ECmdDestroyMixer );
  iShared.iAliveMutex.Wait();
  iShared.iAliveMutex.Close();
  iShared.iMutex.Close();
  if(iShared.iModule != NULL)
  {
	  Player_Free(iShared.iModule);
	  iShared.iModule = NULL;
  }
  MikMod_UnregisterAllDrivers();
  MikMod_Exit();
  CloseSTDLIB();
}


void CSndMixer::Pause() {
  if( iPaused )
    {
    return;
    }
  iPaused = ETrue;
  SendCmd( ECmdStopMixer );
}


void CSndMixer::Resume() {    
  if( !iPaused )
    {
    return;
    }   
  iPaused = EFalse;
  SendCmd( ECmdStartMixer );
}
  

void CSndMixer::Play( const TSample& aSample, TInt aChannel, TInt aFrequency, TInt aVolume ) {
  iShared.iMutex.Wait();
  iShared.iPlayStarted[ aChannel ] = ETrue;
  iShared.iSample[ aChannel ] = aSample;
  iShared.iFrequency[ aChannel ] = aFrequency;
  iShared.iVolume[ aChannel ] = aVolume;
  iShared.iMutex.Signal();
}

/** 
  * Starts music in mikmod
  * @param name of music to load
  */
void CSndMixer::PlayMusic(const char* aFileName) {
  iShared.iMutex.Wait();
  
  if(iShared.iModule != NULL) {
	  Player_Free(iShared.iModule);
	  iShared.iModule = NULL;
  }

  iShared.iModule = Player_Load((CHAR*) aFileName, 20, false);

  if(iShared.iModule != NULL) {
	  Player_Start(iShared.iModule);
  }

  iShared.iMutex.Signal();
}


void CSndMixer::PauseMusic(TBool aPauseState) {
	iShared.iModulePaused = aPauseState;
}

void CSndMixer::Stop( TInt aChannel ) {
  iShared.iMutex.Wait();
  iShared.iPlayStarted[ aChannel ] = ETrue;
  iShared.iSample[ aChannel ] = iEmptySample;
  iShared.iMutex.Signal();  
}


void CSndMixer::SetVolume( TInt aVolume ) {
  iShared.iMutex.Wait();
  iShared.iMainVolume = aVolume;
  iShared.iMutex.Signal();  
}


TInt CSndMixer::Volume() {
  return iShared.iMainVolume;
}


void CSndMixer::SendCmd( TMixerCmd aCmd ) {
  iShared.iMutex.Wait();
  iShared.iCmd = aCmd;
  iShared.iMutex.Signal();
  iShared.iExc = EExcUserInterrupt;
    
  TRequestStatus* status = iShared.iStatusPtr;
  if(status->Int() == KRequestPending )
    {
    iMixerThread.RequestComplete(status, KErrNone);
    }
}

